# -*- coding: utf-8 -*-
'''
Default values, to be imported elsewhere in Salt code

Do NOT, import any salt modules (salt.utils, salt.config, etc.) into this file,
as this may result in circular imports.
'''

# Default delimiter for multi-level traversal in targeting
DEFAULT_TARGET_DELIM = ':'
